Jump = {}
Jump.hook = {}

local function Copy(org, res)
	if not org then
		return {}
	end

	res = res or {}

	for k, v in pairs(org) do
		if type(v) ~= 'table' then
			res[k] = v
		else
			res[k] = {}
			Copy(v, res[k])
		end
	end

	return res
end

local function Array(value, max)
	value = value or false
	max = max or game('sv_maxplayers')

	local array = {}
	if type(value) == 'table' then
		for i = 1, max do
			array[i] = {}
			Copy(value, array[i])
		end
	else
		for i = 1, max do
			array[i] = value
		end
	end

	return array
end

local isJumping = Array(false)
local pl_GoTo = Array({x = nil, y = nil})

local banneds = {}
function Jump.BanArea(fx, fy, tx, ty)
	table.insert(
		banneds,
		{
			fx = fx,
			fy = fy,
			tx = tx,
			ty = ty
		}
	)
end

local function getPOS(p)
	local dx, dy = 0, 0

	do
		local rot = player(p, 'rot')

		if rot % 45 == 0 and rot % 90 ~= 0 then
			if rot > 0 and rot ~= 225 then
				dx = 1
			else
				dx = -1
			end

			if math.abs(rot) == 45 then
				dy = -1
			else
				dy = 1
			end
		else
			local nRot = math.floor((rot + 135) / 90)
			nRot = (nRot == 0 and 4 or nRot)

			if nRot % 2 == 0 then
				dx = (nRot == 2 and 1 or -1)
			else
				dy = (nRot == 3 and 1 or -1)
			end
		end
	end

	return dx, dy
end

local function isWall(x, y)
	local eProperty = tile(x, y, 'property')
	local eType = entity(x, y, 'type')

	local con1 = 1 <= eProperty and eProperty <= 4
	local con2 = not entity(x, y, 'state') and eType == 25
	local con3 = not entity(x, y, 'state') and eType == 71 and entity(x, y, 'int1') ~= 4
	local con4 = tile(x, y, 'frame') == 0

	if con1 or con2 or con3 or con4 then
		return true
	end

	return false
end

local function canJump(x, y)
	for _, banArea in pairs(banneds) do
		if banArea.fx <= x and x <= banArea.tx and banArea.fy <= y and y <= banArea.ty then
			return
		end
	end

	local eType
	if entity(x, y, 'exists') and not entity(x, y, 'state') then
		eType = entity(x, y, 'type')

		-- 30 - Env_Building
		local con1 = eType == 30 and (entity(x, y, 'int0') ~= 1 or objectat(x, y, 1) == 0)

		-- 71 - Func_DynWall
		local con2 = eType == 71 and (entity(x, y, 'int1') ~= 1 and entity(x, y, 'int1') ~= 3)

		-- 25 - Env_Breakable
		local con3 = eType == 25

		if con1 or con2 or con3 then
			return false
		end
	end

	if objectat(x, y) == 1 and objectat(x, y, 1) == 0 then
		return false
	end

	return tile(x, y, 'property') == 2 or tile(x, y, 'property') == 4 or
		(objectat(x, y) == 1 and objectat(x, y, 1) == 1) or
		(eType == 71) or
		(eType == 30 and entity(x, y, 'int0') == 1)
end

local function isTele(x, y)
	if entity(x, y, 'exists') and not entity(x, y, 'state') then
		local eType = entity(x, y, 'type')

		return eType == 70
	end

	return false
end

local function getMathSymbol(a)
	if a > 0 then
		return 1
	elseif a < 0 then
		return -1
	else
		return 0
	end
end

local function ToTableStr(...)
	local _t = {...}
	local str = '{'

	for _, v in pairs(_t) do
		str = str .. v .. ','
	end

	str = str .. 'nil}'

	return str
end

local floors = {
	[10] = 'dirt',
	[11] = 'snow',
	[12] = 'step',
	[13] = 'tile',
	[14] = 'wade',
	[15] = 'metal',
	[16] = 'wood'
}

function Jump.Done(p)
	if not isJumping[p] then
		return
	end

	isJumping[p] = false
	timer(250, 'parse', 'speedmod ' .. p .. ' 0', 1)

	local x = player(p, 'x')
	local y = player(p, 'y')
	local eProperty = tile(math.floor(x / 32), math.floor(y / 32), 'property')

	if 10 <= eProperty and eProperty <= 16 then
		parse('sv_soundpos "player/pl_' .. floors[eProperty] .. math.random(1, 4) .. '.wav" ' .. x .. ' ' .. y)
	end
end

local step = 2
function Jump.Goto(str)
	local _t = loadstring('return ' .. str)()
	local p, gx, gy = unpack(_t)

	if not isJumping[p] then
		return
	end

	if gx ~= 0 or gy ~= 0 then
		local nx = math.abs(gx) > step and step * getMathSymbol(gx) or gx
		local ny = math.abs(gy) > step and step * getMathSymbol(gy) or gy

		local x = player(p, 'x')
		local y = player(p, 'y')

		parse('setpos ' .. p .. ' ' .. x + nx .. ' ' .. y + ny)
		local nxTile = math.floor((x + nx) / 32)
		local nyTile = math.floor((y + ny) / 32)

		if isTele(nxTile, nyTile) then
			parse(
				'setpos ' ..
					p ..
						' ' ..
							entity(nxTile, nyTile, 'int0') * 32 + 16 .. ' ' .. entity(nxTile, nyTile, 'int1') * 32 + 16
			)

			Jump.Done(p)

			return
		end

		timer(500 * step / 32, 'Jump.Goto', ToTableStr(p, gx - nx, gy - ny), 1)
	else
		Jump.Done(p)
	end
end

function Jump.Try(p)
	if player(p, 'process') == 1 or not (player(p, 'exists') and player(p, 'health') > 0 and player(p, 'speedmod') == 0) then
		return
	end

	local dx, dy = getPOS(p)
	local x, y = player(p, 'tilex'), player(p, 'tiley')
	local wx, wy = x + dx, y + dy

	if isTele(x, y) then
		return
	end

	if canJump(wx, wy) and not isWall(wx + dx, wy + dy) then
		parse('speedmod ' .. p .. ' ' .. -100)

		isJumping[p] = true

		pl_GoTo[p].x = wx + dx
		pl_GoTo[p].y = wy + dy

		Jump.Goto(ToTableStr(p, dx * 64, dy * 64))
	end
end

function Jump.hook.use(p, event)
	if event == 0 then
		Jump.Try(p)
	end
end

function Jump.hook.drop(p)
	if isJumping[p] then
		return 1
	end
end

function Jump.hook.die(p)
	if isJumping[p] and pl_GoTo[p].x and pl_GoTo[p].y then
		isJumping[p] = false
		parse('setpos ' .. p .. ' ' .. pl_GoTo[p].x * 32 + 16 .. ' ' .. pl_GoTo[p].y * 32 + 16)
	end
end

function Jump.hook.startround_prespawn()
	for i = 1, 32 do
		parse('speedmod ' .. i .. ' ' .. 0)
		isJumping[i] = false
		pl_GoTo[i].x = nil
		pl_GoTo[i].y = nil
	end
end
Jump.hook.startround_prespawn()

for funcName in pairs(Jump.hook) do
	addhook(funcName, 'Jump.hook.' .. funcName)
end
